Attribute VB_Name = "basAiLoad"
Option Explicit

Const VERSION_LOOK& = -1
Const VERSION_NONE& = -2
Const VERSION_LOST& = -3
Const VERSION_FAIL& = -4

Dim port&
Dim baud$
Dim filename$
Dim bootloadCommand&

Dim version&
Dim LFSR&
Dim txFifo$
Dim TxBitFifo$
Dim rxFifo$

Dim zroBit$ ' "001" or "001"
Dim oneBit$ ' "011" or "01"

Dim txBytesRead&
Dim txBytesSent&
Dim txTimeTaken&

Private Declare _
  Function GetTickCount _
  Lib "kernel32" () As Long
Sub Failed(ByVal errorMessage$)
  MsgBox errorMessage$, vbExclamation, frmAiLoad.Caption
  If version& >= 0 Then
    version& = VERSION_FAIL
  End If
End Sub
Sub Okay(ByVal okayMessage$)
  MsgBox okayMessage$, vbInformation, frmAiLoad.Caption
End Sub

Function FNdir$(ByVal f$)
Dim result$
  result$ = ""
  On Error Resume Next
  result$ = Dir$(f$)
  On Error GoTo 0
  FNdir$ = result$
End Function

Function FNhex$(ByVal n&, ByVal w&)
  FNhex$ = Right$(String$(w&, "0") + Hex$(n&), w&)
End Function

Function FNpercent&(ByVal n&, ByVal total&)
  FNpercent& = Int((n& / total&) * 100)
End Function

Sub ImportSettings()
  With frmAiLoad
    .barProgress.Value = 0
    port& = Val(Trim$(.txtPort.Text))
    baud$ = Trim$(.txtBaud.Text)
    filename$ = Trim$(.txtFilename.Text)
    If .chkDensePacking.Value Then
      zroBit$ = "001"
      oneBit$ = "01"
    Else
      zroBit$ = "001"
      oneBit$ = "011"
    End If
    If .chkWriteEeprom.Value = 0 Then
      bootloadCommand& = 1
    Else
      bootloadCommand& = 3
    End If
  End With
End Sub

Sub RxdData(ByVal rxData$)
  rxFifo$ = rxFifo$ + rxData$
End Sub

Function RxRawBit&()
Dim bit&
  Select Case RxRawByte&()
    Case &HFE: bit& = 0
    Case &HFF: bit& = 1
    Case Else: bit& = -1
  End Select
  RxRawBit& = bit&
End Function


Function RxBitResponse&(ByVal timeout&)
  RxBitResponse& = RxBitSolicitOrResponse&(timeout&, False)
End Function
Function RxBitSolicit&(ByVal timeout&)
  RxBitSolicit& = RxBitSolicitOrResponse&(timeout&, True)
End Function
Sub SerialPort_Close()
  With frmAiLoad.SerialPort
    If .PortOpen Then
      .PortOpen = False
    End If
  End With
  frmAiLoad.btnClosePort.Enabled = False
End Sub


Sub TxBit(ByVal bit&)
  If bit& = 0 Then
    Call TxAddBit(0)
  Else
    Call TxAddBit(1)
  End If
End Sub

Sub TxByte(ByVal b&)
Dim i&
  For i& = 0 To 7
    If (b& And (2 ^ i&)) = 0 Then
      Call TxAddBit(0)
    Else
      Call TxAddBit(1)
    End If
  Next
End Sub

Sub TxBitFifoFlush()
  
  ' Make sure the txBitFifo bit stream is padded out to full bytes and
  ' put into the txFifo byte stream.
  
  If Len(TxBitFifo$) <> 0 Then
    If Len(TxBitFifo$) < 9 Then
      TxBitFifo$ = TxBitFifo$ + String$(9 - Len(TxBitFifo$), "1")
    End If
    Call TxAddBit(-1)
  End If

End Sub

Sub TxAddBit(bit&)
Dim tx$
Dim i&
Dim b&
  Select Case bit&
    Case 0: TxBitFifo$ = TxBitFifo$ + zroBit$ ' 001 or 001
    Case 1: TxBitFifo$ = TxBitFifo$ + oneBit$ ' 011 or 01
  End Select
  While Len(TxBitFifo$) >= 9
    ' Possibilities if we are here are ...
    '   S01234567E      Send            Leaves
    '   0xxxx1001       xxxx1001        <nothing>
    '   0xxxxx101       xxxxx101        <nothing>
    '   0xxxxx1001...   xxxxx100        ...
    '   0xxxxxx101...   xxxxxx10        ...
    '   0xxxxxx1001...  xxxxxx11        001...
    If Len(TxBitFifo$) = 9 Then
      tx$ = Mid$(TxBitFifo$, 2, 8)
      TxBitFifo$ = ""
    Else
      ' Check if we have a 2T extending over the byte boundary
      If Mid$(TxBitFifo$, 9, 2) = "00" Then
        ' We have a 2T extending over the byte boundary. We need to transmit
        ' what we have before the 2T and leave the 2T for the next byte
        tx$ = Mid$(TxBitFifo$, 2, 7) + "1"
        TxBitFifo$ = Right$(TxBitFifo$, Len(TxBitFifo$) - 8)
      Else
        tx$ = Mid$(TxBitFifo$, 2, 8)
        TxBitFifo$ = Right$(TxBitFifo$, Len(TxBitFifo$) - 10)
      End If
    End If
    ' Change the tx$ into a byte value. Note the leading zero has already
    ' been stripped away which will be the start bit.
    b& = 0
    For i& = 0 To 7
      b& = b& \ 2
      If Mid$(tx$, i& + 1, 1) = "1" Then
        b& = b& Or &H80
      End If
    Next
    ' Add the byte to the outgoing byte stream
    txFifo$ = txFifo$ + Chr$(b&)
  Wend
End Sub


Sub TxSync()
  Call TxRawByte(&HF9)
End Sub
Sub TxRawBit(ByVal b&)
  Call TxRawByte(&HFE Or (b& And 1))
End Sub




Sub Do_Identify()
  Call SerialPort_Open(port&, baud$)
  Select Case version&
    Case VERSION_NONE: Call Failed("No Propeller Chip found")
    Case VERSION_LOST: Call Failed("Hardware link lost")
    Case Else: Call Okay("Propeller Chip Version" + Str$(version&) + " found")
  End Select
End Sub
Sub Do_Download()
Const MAX_BLOCK_SIZE& = 128
Dim block$
Dim blockSize&
Dim fileNumber&
Dim i&

  If Trim$(filename$) = "" Then
    Call Failed("No file selected")
  Else
    If FNdir$(filename$) = "" Then
      Call Failed("File not found" + vbCrLf + vbCrLf + filename$)
    Else
      Call SerialPort_Open(port&, baud$)
      Select Case version&
        Case VERSION_NONE: Call Failed("No Propeller Chip found")
        Case VERSION_LOST: Call Failed("Hardware link lost")
        Case Else
          fileNumber& = FreeFile
          Open filename$ For Binary As #fileNumber&
          If (LOF(fileNumber&) Mod 4) <> 0 Then
            Call Failed("Invalid file size")
          Else
  
            txBytesRead& = LOF(fileNumber&) + 8
            txBytesSent& = 0
            txTimeTaken& = GetTickCount
  
            ' Send download command
  
            Call TxByte(bootloadCommand&)  ' 1=Ram, 3=Eeprom
            Call TxByte(0)
            Call TxByte(0)
            Call TxByte(0)
            Call TxSend
  
            ' Send size in longs
  
            i& = LOF(fileNumber&) \ 4
  
            Call TxByte(i& And &HFF)
            Call TxByte(i& \ &H100)
            Call TxByte(0)
            Call TxByte(0)
            Call TxBitFifoFlush
  
            ' Send the image
  
            Do
              frmAiLoad.barProgress.Value = FNpercent(Loc(fileNumber&), LOF(fileNumber&))
              blockSize& = LOF(fileNumber&) - Loc(fileNumber&)
              If blockSize& > 0 Then
                If blockSize& > MAX_BLOCK_SIZE Then
                  blockSize& = MAX_BLOCK_SIZE
                End If
                block$ = Space$(blockSize&)
                Get #fileNumber&, , block$
                If oneBit$ = "011" Then
                  ' 3-bits per byte packing
                  For i& = 1 To blockSize& Step 4
                    Call TxByte(Asc(Mid$(block$, i& + 0, 1)))
                    Call TxByte(Asc(Mid$(block$, i& + 1, 1)))
                    Call TxByte(Asc(Mid$(block$, i& + 2, 1)))
                    Call TxByte(Asc(Mid$(block$, i& + 3, 1)))
                    Call TxBitFifoFlush
                  Next
                Else
                  ' Dense packing
                  For i& = 1 To blockSize&
                    Call TxByte(Asc(Mid$(block$, i&, 1)))
                  Next
                End If
                Call TxPartSend
              End If
            Loop Until blockSize& = 0
          End If
          Close #fileNumber&
          If version& >= 0 Then
            Call TxSend
            txTimeTaken& = GetTickCount - txTimeTaken&
            rxFifo$ = ""
            If RxBitSolicit&(250) <> 0 Then
              If version& < 0 Then
                Call Failed("Timedout waiting for Checksum Okay response")
              Else
                Call Failed("Checksum failed")
              End If
            Else
              If bootloadCommand& > 1 Then
                If RxBitSolicit&(5000) <> 0 Then
                  If version& < 0 Then
                    Call Failed("Timedout waiting for Write Eeprom Okay response")
                  Else
                    Call Failed("Write Eeprom Failed")
                  End If
                Else
                  If RxBitSolicit&(5000) <> 0 Then
                    If version& < 0 Then
                      Call Failed("Timedout waiting for Verify Eeprom Okay response")
                    Else
                      Call Failed("Verify Eeprom Failed")
                    End If
                  Else
                    Call Okay("Download to Eeprom successful - Sent" + Str$(txBytesSent&) + " bytes")
                  End If
                End If
              Else
                Call Okay("Download to Ram successful - Sent" + Str$(txBytesSent&) + " bytes")
              End If
            End If
          End If
      End Select
    End If
  End If
End Sub

Sub Do_Find()
Dim o$
Dim port&
Dim found&
  found& = 0
  For port& = 1 To 10
    o$ = o$ + vbCrLf + Str$(port&) + " : "
    Call SerialPort_Open(port&, baud$)
    If version& >= 0 Then
      o$ = o$ + "Propeller Chip Version" + Str$(version&) + " found"
      If found& = 0 Then
        found& = port&
      Else
        found& = -1
      End If
    Else
      o$ = o$ + "-"
    End If
  Next
  Call SerialPort_Close
  If found& = 0 Then
    o$ = "No Propeller Chips found" + vbCrLf + o$
    Call Failed(o$)
  Else
    If found& > 0 Then
      o$ = "One Propeller Chip found" + vbCrLf + o$
      frmAiLoad.txtPort.Text = LTrim$(Str$(found&))
    Else
      o$ = "Multiple Propeller Chips found" + vbCrLf + o$
    End If
    Call Okay(o$)
  End If
End Sub

Sub Do_ShutDown()
  Call SerialPort_Open(port&, baud$)
  Select Case version&
    Case VERSION_NONE: Call Failed("No Propeller Chip found")
    Case VERSION_LOST: Call Failed("Hardware link lost")
    Case Else
      Call TxByte(0)  ' 0=Shutdown
      Call TxByte(0)
      Call TxByte(0)
      Call TxByte(0)
      Call TxSend
      Call Okay("Shutdown Sent")
  End Select
End Sub

Sub Do_Reboot()
  Call SerialPort_Open(port&, baud$)
  Select Case version&
    Case VERSION_NONE: Call Failed("No Propeller Chip found")
    Case VERSION_LOST: Call Failed("Hardware link lost")
    Case Else: Call Okay("Rebooted")
  End Select
End Sub

Sub SerialPort_Open(ByVal port&, ByVal baud$)
Dim i&
Dim versionNumber&

  ' Opens the port, checks LFSR and version. Sets version& >=0 if okay
  ' or sets a negative error value.
  
  version& = VERSION_LOOK

  With frmAiLoad.SerialPort
    If .CommPort <> port& Or Left$(.Settings, InStr(.Settings, ",") - 1) <> baud$ Then
      Call SerialPort_Close
    End If
    If Not .PortOpen Then
      On Error Resume Next
      .CommPort = port&
      .Settings = baud$ + ",n,8,1"
      .PortOpen = True
      On Error GoTo 0
      If .PortOpen Then
        frmAiLoad.btnClosePort.Enabled = True
        frmAiLoad.txtPort.Text = LTrim$(Str$(.CommPort))
        frmAiLoad.txtBaud.Text = Left$(.Settings, InStr(.Settings, ",") - 1)
      Else
        version& = VERSION_NONE
      End If
    End If
  End With

  If version& = VERSION_LOOK Then

    txFifo$ = ""
    TxBitFifo$ = ""
    rxFifo$ = ""

    txBytesRead& = 0
    txBytesSent& = 0
    txTimeTaken& = 0

    Call TxReset

    Call TxSync
    Call TxSend

    LFSR = Asc("P")
    If oneBit$ = "011" Then
      For i& = 1 To 250
        Call TxRawBit(IterateLFSR)
      Next
    Else
      For i& = 1 To 250
        Call TxBit(IterateLFSR)
      Next
    End If

    For i& = 1 To 250 + 8
      Call TxSync
    Next

    Call TxSend

    rxFifo$ = ""
    i& = 1
    While i& <= 250 And version& = VERSION_LOOK&
      If RxBitResponse&(100) <> IterateLFSR Then
        version& = VERSION_NONE
      End If
      i& = i& + 1
    Wend

    versionNumber& = 0
    i& = 1
    While i& <= 8 And version& = VERSION_LOOK&
      versionNumber& = (RxBitResponse&(50) * &H80) Or (versionNumber& / 2)
      i& = i& + 1
    Wend

    If version& = VERSION_LOOK Then
      Select Case versionNumber&
        Case 1:    version& = versionNumber&
        Case Else: version& = VERSION_NONE
      End Select
    End If

  End If

End Sub
Sub TxReset()
  With frmAiLoad.SerialPort
    .DTREnable = True
    Call Wait_Ms(15)
    .DTREnable = False
    Call Wait_Ms(100)
  End With
End Sub

Sub Wait_Ms(ByVal msTimeout&)
Dim timeout&

  ' This may need some tweaking if there's a 'past midnight' overflow /
  ' wrap-round in GetTick Count.
  
  timeout& = GetTickCount + msTimeout&
  While GetTickCount < timeout&
    DoEvents
  Wend

End Sub

Function RxBitSolicitOrResponse&(ByVal timeout&, ByVal sendSync As Boolean)
Dim result&

  result& = -1

  While timeout& > 0 And result& < 0
    If sendSync Then
      Call TxSync
      Call TxSend
    End If
    result& = RxRawBit&()
    If result& < 0 Then
      Call Wait_Ms(10)
    End If
    timeout& = timeout& - 10
  Wend

  If result& < 0 Then
    If version& = VERSION_LOOK Then
      version& = VERSION_NONE
    Else
      version& = VERSION_LOST
    End If
    result& = 0
  End If

  RxBitSolicitOrResponse& = result&

End Function

Function RxRawByte&()
Dim result&
  If Len(rxFifo$) = 0 Then
    result& = -1
  Else
    result& = Asc(Left$(rxFifo$, 1))
    rxFifo$ = Right$(rxFifo$, Len(rxFifo$) - 1)
  End If
  RxRawByte& = result&
End Function



Sub TxRawByte(ByVal byt&)
  Call TxBitFifoFlush
  txFifo$ = txFifo$ + Chr$(byt& And &HFF)
End Sub
Sub TxSend()

  Call TxBitFifoFlush

  txBytesSent& = txBytesSent& + Len(txFifo$)

  With frmAiLoad.SerialPort
    While Len(txFifo$) <> 0
      While .OutBufferCount <> 0
        DoEvents
      Wend
      If Len(txFifo$) > .OutBufferSize Then
        .Output = Left$(txFifo$, .OutBufferSize)
        txFifo$ = Right$(txFifo$, Len(txFifo$) - .OutBufferSize)
      Else
        .Output = txFifo$
        txFifo$ = ""
      End If
    Wend
  End With

End Sub
Sub TxPartSend()

  With frmAiLoad.SerialPort
    While Len(txFifo$) >= .OutBufferSize
      While .OutBufferCount <> 0
        DoEvents
      Wend
      .Output = Left$(txFifo$, .OutBufferSize)
      txFifo$ = Right$(txFifo$, Len(txFifo$) - .OutBufferSize)
      txBytesSent& = txBytesSent& + .OutBufferSize
    Wend
  End With

End Sub



Function IterateLFSR&()
Dim result&
Dim newLsb&

  result& = LFSR& And 1

  newLsb& = ((LFSR& And &H80) <> 0) Xor _
            ((LFSR& And &H20) <> 0) Xor _
            ((LFSR& And &H10) <> 0) Xor _
            ((LFSR& And &H2) <> 0)

  LFSR& = ((LFSR& And &H7F) * 2) Or (newLsb& And 1)

  IterateLFSR& = result&

End Function



Private Sub Form_Unload(Cancel As Integer)
  Call SerialPort_Close
End Sub
